/* wipe 0.11
 *
 * by Berke Durak <berke@gsu.linux.org.tr>
 *
 * Homepage for wipe: http://gsu.linux.org.tr/wipe/
 *
 * Securely erase files from magnetic media
 * Based on data from article "Secure Deletion of Data from Magnetic
 * and Solid-State Memory" from Peter Gutmann.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef HAVE_GETOPT
#include <getopt.h>
#endif
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "random.h"
#include "misc.h"

#define DEVRANDOM "/dev/random"

#define OPTSTR "fhvrqspciR:S:M:"

int num_errors = 0;
int num_files = 0;
int num_dirs = 0;
int num_spec = 0;

char *progname;
int o_force = 0;
int o_dochmod = 0;
int o_errorabort = 0;
int o_recurse = 0;
int o_quick = 0;
int o_verbose = 0;
int o_silent = 0;
char *o_devrandom = DEVRANDOM;
int o_randseed = RANDS_DEVRANDOM;
int o_randalgo = RANDA_ARCFOUR;
int o_randseed_set = 0;

#ifdef REMOVE_ON_INT
int doremove = 0;
char toremove[512];
#endif

void interrupt (int i)
{
	fflush (stdout);
	fflush (stderr);
	fprintf (stderr, "\n*** Interrupted by signal %d\n", i);
#ifdef REMOVE_ON_INT
	if (toremove) {
		fprintf (stderr, "*** Removing %s\n", toremove);
		if (remove (toremove)) fprintf (stderr, "*** Error: %s\n", strerror (errno));
	}
#endif
	exit (1);	/* exit () does fflush () */
}

void fill_random (char *b, int n)
{
	u32 l;
	
	for (; n >= 4; n -= 4) {
		l = rand_Get32 ();

		*(b++) = l & 0xff; l >>=8;
		*(b++) = l & 0xff; l >>=8;
		*(b++) = l & 0xff; l >>=8;
		*(b++) = l & 0xff;
		
		n -= 4;
	}

	if (n) {
		l = rand_Get32 ();
		while (n--) {
			*(b++) = l & 0xff; l >>= 8;
		}
	}
}

void fill_pattern (char *b, int n, char *p, int l)
{
	int i, j = 0;

	for (i = 0; i<n; i++) {
		*(b++) = p[j];
		j ++;
		if (j == l) j = 0;
	}
}

/* P.Gutmann, in his article, was recommening to make
 * the DETERMINISTIC passes (i.e. those that don't write random data)
 * in random order, and to make 4 random passes before and after these. 
 *
 * Here, the deterministic and the 8 random passes are made in random
 * order, unless, of course, the "quick" mode is selected.
 */

#define BUFLG2 13
#define BUFSIZE (1<<BUFLG2)

struct passinfo_s {
	int type;
	int len;
	char *pat;
} passinfo[] = {
	{ 0, 0, 0,		},	/* 1  random */
	{ 0, 0, 0,		},	/* 2  random */
	{ 0, 0, 0,		},	/* 3  random */
	{ 0, 0, 0, 		},	/* 4  random */
	{ 1, 1, "\x55", 	},	/* 5  RLL MFM */
	{ 1, 1, "\xaa", 	},	/* 6  RLL MFM */
	{ 1, 3, "\x92\x49\x24", },	/* 7  RLL MFM */
	{ 1, 3, "\x49\x24\x92", },	/* 8  RLL MFM */
	{ 1, 3, "\x24\x92\x49", },	/* 9  RLL MFM */
	{ 1, 1, "\x00", 	},	/* 10 RLL */
	{ 1, 1, "\x11", 	},	/* 11 RLL */
	{ 1, 1, "\x22", 	},	/* 12 RLL */
	{ 1, 1, "\x33", 	},	/* 13 RLL */
	{ 1, 1, "\x44", 	},	/* 14 RLL */
	{ 1, 1, "\x55", 	},	/* 15 RLL */
	{ 1, 1, "\x66", 	},	/* 16 RLL */
	{ 1, 1, "\x77", 	},	/* 17 RLL */
	{ 1, 1, "\x88", 	},	/* 18 RLL */
	{ 1, 1, "\x99", 	},	/* 19 RLL */
	{ 1, 1, "\xaa", 	},	/* 20 RLL */
	{ 1, 1, "\xbb", 	},	/* 21 RLL */
	{ 1, 1, "\xcc", 	},	/* 22 RLL */
	{ 1, 1, "\xdd", 	},	/* 23 RLL */
	{ 1, 1, "\xee", 	},	/* 24 RLL */
	{ 1, 1, "\xff", 	},	/* 25 RLL */
	{ 1, 3, "\x92\x49\x24",	},	/* 26 RLL MFM */
	{ 1, 3, "\x49\x24\x92",	},	/* 27 RLL MFM */
	{ 1, 3, "\x24\x92\x49",	},	/* 28 RLL MFM */
	{ 1, 3, "\x6d\xb6\xdb",	},	/* 29 RLL */
	{ 1, 3, "\xb6\xdb\x6d",	},	/* 30 RLL */
	{ 1, 3, "\xdb\x6d\xb6",	},	/* 31 RLL */
	{ 0, 0, 0,		},	/* 32 random */
	{ 0, 0, 0,		},	/* 33 random */
	{ 0, 0, 0,		},	/* 34 random */
	{ 0, 0, 0, 		},	/* 35 random */
};

#define MAXPASSES (sizeof(passinfo)/(sizeof(*passinfo)))
#define QUICKPASSES 4
#define fnerror(x)  { num_errors++; fprintf (stderr, "%.32s: " x ": %.32s\n", fn, strerror (errno)); }
#define fnerrorq(x) { num_errors++; fprintf (stderr, "%.32s: " x "\n", fn); }

int dothejob (char *fn)
{
	int fd;
	int p[MAXPASSES];

	int bpi = 0;	/* block progress indicator enabled ? */
	int i;
	off_t j;
	char buf[BUFSIZE];
	struct stat st;
	struct passinfo_s *pi;
	off_t l;
	time_t lt = 0, t;

	/* to do a cryptographically strong random permutation on the 
	 * order of the passes, we need lg_2(MAXPASSES!) bits of entropy:
	 * for MAXPASSES=35, this means about 132 bits.
	 *
	 * rand_seed () should give 31 bits of entropy per call
	 * we need a maximum of lg_2(MAXPASSES) bits of entropy per random
	 * value. therefore we should call rand_seed () at least for
	 * every 31/lg_2(MAXPASSES) ~= 6 random values.
	 *
	 * on the other way, as we'll constantly force disk access,
	 * /dev/urandom should have no difficulty providing us good random
	 * values.
	 */

	if (!o_quick) {
		for (i = 0; i<MAXPASSES; p[i]=i, i++);

		for (i = 0; i<MAXPASSES-2; i++) {
			int a, b;

			/* a \in { 0, 1, ... MAXPASSES-i-1 } */

			/* since MAXPASSES-i is not necessarily a divisor of
			 * 2^32, we won't get uniform distribution
			 * with this. however, since MAXPASSES is very
			 * small compared to RAND_MAX, ti
			 */
			
			a = rand_Get32 () % (MAXPASSES-i);
			b = p[i+a];
			p[i+a] = p[i];
			p[i] = b;
		}
	}
	
	/* see what kind of file it is */
	
	if (lstat (fn, &st)) { fnerror("stat error"); return -1; }

	if (S_ISDIR(st.st_mode)) {
		if (!o_silent) fprintf (stderr, "Skipping directory %.32s...\n", fn);
		return 1;
	}

	if (S_ISREG(st.st_mode)) {
#ifdef HAVE_OSYNC
	 	fd = open (fn, O_WRONLY | O_SYNC);
#else
		fd = open (fn, O_WRONLY);
#endif
	 	if (fd < 0) {
	 		if (errno == EACCES) {
	 			if (o_force || o_dochmod) {
	 				if (chmod (fn, 0700)) {
	 					fnerror("chmod error");
	 					return -1;
	 				} else fd = open (fn, O_WRONLY);
	 			} else { fnerror("permission error: try -c"); return -1; }
	 		} else { fnerror("open error"); return -1; }
	 	}
	 	if (fd < 0) { fnerror("open error even with chmod"); return -1; }
	 	
#ifdef REMOVE_ON_INT
	 	strcpy (toremove, fn);
	 	doremove = 1;
#endif
	 	
	 	l = (st.st_size>>BUFLG2) + 1;

		/* do the passes */

	 	for (i = 0; i<(o_quick?QUICKPASSES:MAXPASSES); i++) {
	 		if (!o_silent) {
		 		if (o_quick) 
			 		fprintf (stderr, "\rWiping %.32s, pass %d in quick mode ", fn, i);
				else
		 			fprintf (stderr, "\rWiping %.32s, pass %-2d (%-2d) ", fn, i, p[i]);
			}
			
	 		lseek (fd, 0, SEEK_SET);

	 		pi = passinfo + p[i];
			if (!o_quick && pi->type)
				fill_pattern (buf, BUFSIZE, pi->pat, pi->len);

			if (!o_silent) lt = time (0);

	 		for (j = 0; j<l; j ++) {
	 			if (!o_silent) {
	 				t = time (0);
	 				if ((bpi && (t-lt)) || ((t-lt>2) && j<(l>>1))) {
	 					fprintf (stderr,
	 						"[%5ld / %5ld]\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", (long) j, (long) l);
	 					fflush (stderr);
						lt = t;
						bpi = 1;
	 				}
	 			}
				if (o_quick || !pi->type) {
	 				/* rand_seed ();	*/		/* cryptographic seed */
	 				fill_random (buf, BUFSIZE);
	 			}
	 			write (fd, buf, BUFSIZE);
#ifndef HAVE_OSYNC
				if (fsync (fd)) {
					fnerror ("fsync error [1]")
					close (fd);
					return -1;
				}
#endif
	 		}

	 		if (fsync (fd)) {
	 			fnerror ("fsync error [2]");
	 			close (fd);
	 			return -1;
	 		}
	 	}
	 	close (fd);
	}
	
	if (remove (fn)) { fnerror("remove"); return -1; }
#ifdef REMOVE_ON_INT
	doremove = 0;
#endif
	if (!o_silent) fprintf (stderr, "\r                                                                          \r");
	if (o_verbose) printf ("File %.32s (%ld bytes) wiped\n", fn, (long) st.st_size);
	num_files ++;
	return 0;
}

int recursive (char *fn)
{
	int r = 0;
	struct stat st;
	char *olddir;
	
	if (lstat (fn, &st)) { fnerror ("stat error"); return -1; }
	
	if (S_ISDIR(st.st_mode)) {
		DIR *d;
		struct dirent *de;
		
		olddir = getcwd (0, 4096);
		if (!olddir) { fnerror("getcwd"); return -1; }

		if (o_verbose) printf ("Entering directory '%s'\n", fn);

		/* fix on 14.02.1998 -- check r/w/x permissions on directory, fix if necessary */
		if ((st.st_mode & 0700) != 0700) {
			if (o_dochmod) {
				if (o_verbose) printf ("Changing permissions from %04o to %04o\n",
							(int) st.st_mode, (int) st.st_mode | 0700);
				if (chmod (fn, st.st_mode | 0700)) {
					fnerror ("chmod [1]");
					return -1;
				}
			} else {
				fnerrorq ("directory mode is not u+rwx; use -c option");
				return -1;
			}
		}

		d = opendir (fn);
		if (!d) {
			if (errno == EACCES && o_dochmod) {
				if (chmod (fn, st.st_mode | 0700)) {
					fnerror("chmod [2]"); return -1;
				} else d = opendir (fn);
			} else { fnerror("opendir"); return -1; }
		}
		if (!d) { fnerror("opendir after chmod"); return -1; }

		if (chdir (fn)) { fnerror("chdir"); return -1; }

		errno = 0;
		num_dirs ++;

		while ((de = readdir (d))) {
			if (strcmp (de->d_name, ".") && strcmp (de->d_name, "..")) {
				if (recursive (de->d_name)) {
					/* num_errors ++; */
					r = -1;
					if (o_errorabort) break;
				}
			}
			errno = 0;
		}
		
		if (errno) { fnerror("readdir"); return -1; }
		closedir (d);
		if (o_verbose) printf ("Going back to directory %s\n", olddir);
		if (chdir (olddir)) { fnerror("chdir .."); return -1; }
		free (olddir);
		if (!r && rmdir (fn)) { fnerror ("rmdir"); return -1; }	
	} else {
		if (S_ISREG(st.st_mode)) {
			return dothejob (fn);
		} else num_spec ++;
		if (remove (fn)) { fnerror ("remove"); return -1; }	
	}

	return r;
}

void banner ()
{
	fprintf (stderr, "wipe version 0.04, 06.08.1998 by Berke Durak <berke@gsu.linux.org.tr>\n"
"\nBased on data from \"Secure Deletion of Data from Magnetic and Solid-State\n"
"Memory\" by Peter Gutmann <pgut001@cs.auckland.ac.nz>\n");
}

void usage (char *msg)
{
	if (msg) fprintf (stderr, "Syntax error: %s\n", msg);
	
	fprintf (stderr,
		"Usage:\t\t%s [options] files...\n"
		"Options:\t-f Force, ie. don't ask for confirmation\n"
		"\t\t-c Do chmod on write-protected files\n"
		"\t\t-r Recurse into directories\n"
		"\t\t-q Quick wipe, less secure, 4 random passes\n"
		"\t\t-a Abort on error\n"
		"\t\t-i Informational (verbose) mode\n"
		"\t\t-s Silent mode\n"
		"\t\t-R set random device OR random seed command\n"
		"\t\t-S (r|c|p) random seed method\n"
		"\t\t\t r: read from random device (strong)\n"
		"\t\t\t c: read from output of random seed command\n"
		"\t\t\t p: use pid (), clock () etc. (weakest)\n"
		"\t\t-M (l|r) set PRNG algorythm\n"
		"\t\t\t l: use libc ()'s "
#ifdef HAVE_RANDOM
		"random ()"
#else
		"rand ()"
#endif
					"library call\n"
#ifdef RC6_ENABLED
		"\t\t\t r: use RC6 encryption algorythm\n"
#endif
		"\t\t\t a: use arcfour encryption algorythm\n"
		"\t\t-v Show version information\n"
		"\t\t-h Display this help\n",
		progname
	);
	exit (EXIT_FAILURE);
}

int main (int argc, char **argv)
{
	int i;
	int n, ndir, nreg;
	int c;
	struct stat st;
	
	/* basic setup */
	
	progname = *argv;
	signal (SIGINT,  interrupt);
	signal (SIGTERM, interrupt);

	/* parse options */
	
	for (;;) {
		c = getopt (argc, argv, OPTSTR);
		if (c<0) break;

		switch (c) {
			case 'f':	o_force = 1;			break;
			case 'v':	banner ();				exit (0);
			case 'r':	o_recurse = 1;			break;
			case 'q':	o_quick = 1;			break;
			case 'c':	o_dochmod = 1;			break;
			case 'i':	o_verbose = 1;			break;
			case 's':	o_silent = 1;			break;
			case 'R':	o_devrandom = optarg;
					o_randseed_set = 1;
					break;
			case 'S':
				if (optarg[1]) 
					usage ("'random seed method' must be single char");
				switch (optarg[0]) {
					case 'r':
						o_randseed = RANDS_DEVRANDOM; break;
					case 'c':
						o_randseed = RANDS_PIPE; break;
					case 'p':
						o_randseed = RANDS_PID; break;
					default:
						usage ("unknown random seed method, must be r,c or p");
						break;
				}
				o_randseed_set = 1;
				break;
			case 'M':
				if (optarg[1])
					usage ("'PRNG algorythm' must be single char");
				switch (optarg[0]) {
					case 'l':
						o_randalgo = RANDA_LIBC; break;
#ifdef RC6_ENABLED
					case 'r':
						o_randalgo = RANDA_RC6; break;
#endif
					case 'a':
						o_randalgo = RANDA_ARCFOUR; break;
					default:
						usage ("unknown random seed method, see help");
						break;
				}
				break;
			case 'h':
			case '?':
			default:
				usage (0);
				/* not reached */
		}
	}
	
	if (optind >= argc) usage ("wrong number of arguments");
	
	/* automatic detection of a suitable random device */

	if (!o_randseed_set) {
		char *t;
		
		t = getenv ("WIPE_SEEDPIPE");
		
		if (t || access (o_devrandom, R_OK)) {
			o_devrandom = t;
			if (access (o_devrandom, X_OK))
				o_randseed = RANDS_PID;
			else o_randseed = RANDS_PIPE;
		}
	}

	/* initialise PRNG */
	rand_Init ();

	/* stat specified files/directories */
	
	n = argc-optind;
	ndir = nreg = 0;
	
	for (i = optind; i<argc; i++) {
		if (lstat (argv[i], &st)) {
			fprintf (stderr, "%s: fatal: could not lstat: %s\n",
					argv[i], strerror (errno));
			exit (1);
		}
		if (S_ISDIR (st.st_mode)) ndir ++;
		else if (S_ISREG (st.st_mode)) nreg ++;
	}
	
	if (!o_recurse && ndir) {
		if (!o_silent && (n - ndir)) fprintf (stderr, "Warning - will skip %d director%s\n", ndir, ndir>1?"ies":"y");
		else {
			fprintf (stderr, "Use -r option to wipe directories\n");
			exit (1);
		}
	} 
	
	if (!o_force) {	
		char buf2[80];
		char buf[80];
		char *b = buf;
		
		if (nreg) {
			sprintf (b, "%d regular file%s", nreg, (nreg>1)?"s":"");
			b += strlen (b);
		}

		if (ndir && o_recurse) {
			sprintf (b, "%s%d director%s",
					(b != buf)?((n-nreg-ndir)?", ":" and "):"",
					ndir, ndir>1?"ies":"y");
			b += strlen (b);
		}
		
		if (n-nreg-ndir) {
			sprintf (b, "%s%d special file%s",
					(b != buf)?" and ":"",
					n-nreg-ndir,
					(n-nreg-ndir)>1?"s":"");
		}
		
		for (;;) {
			fprintf (stderr, "Okay to WIPE %s ? (Yes/No) ", buf);
			fflush (stderr);
			fgets (buf2, sizeof (buf2), stdin);
			if (buf2[0]) buf2[strlen (buf2)-1] = 0;

#ifdef HAVE_STRCASECMP
			if (!strcasecmp (buf2, "no"))
#else
			if ('n' == tolower (buf2[0]) && 'o' == tolower (buf2[1]) && !tolower (buf2[2]))
#endif
			{
				fprintf (stderr, "Aborted.\n");
				exit (0);
			}

#ifdef HAVE_STRCASECMP
			if (strcasecmp (buf2, "yes"))
#else
			if ('y' != tolower (buf2[0]) ||
			    'e' != tolower (buf2[1]) ||
			    's' != tolower (buf2[2]) ||
			    tolower (buf2[3]))
#endif
			{
				fprintf (stderr, "Please answer \"Yes\" or \"No\"\n");
			} else break;
		}
	}
	
	for (i = optind; i<argc; i++) {
		int r;
	
		if (o_recurse) r = recursive (argv[i]);
		else r = dothejob (argv[i]);

#ifdef REMOVE_ON_INT
		doremove = 0;
#endif
		
		/* if (r < 0) num_errors ++; */
	}

	if (!o_silent) fprintf (stderr, "Syncing..."); fflush (stderr);
	sync (); sleep (1); sync ();
	if (!o_silent) fprintf (stderr, "\rOperation finished.\n"
					"%d file%s (%d special) in %d director%s wiped, "
					"%d error%s occured\n",
					num_files+num_spec, (1==num_files+num_spec)?"":"s",
					num_spec,
					num_dirs, (1==num_dirs)?"y":"ies",
					num_errors, (1==num_errors)?"":"s");
	
	return 0;
}
